﻿Public Class FormArrayData
    Private m_OASConfiguration As New OPCSystems.OPCSystemsComponent
    Private WithEvents m_OASData As New OPCSystemsDataConnector.OPCSystemsData
    Private m_OASDataNetwork As New OPCSystemsDataConnector.OPCSystemsDataNetworkNodes
    Private m_InSimuation As Boolean = False
    Private m_LastNetworkNodeUsed As String = "localhost"

    ' Value stack in the controller
    Private m_DateQueue As New Queue
    Private m_ValueQueue1 As New Queue
    Private m_ValueQueue2 As New Queue
    Private m_ValueQueue3 As New Queue
    Private m_LastDate As Date
    Private m_LastTimeFeedbackReceived As Date

    ' Array values to send to OAS
    Private m_Date(999) As String
    Private m_Value1(999) As Double
    Private m_Value2(999) As Double
    Private m_Value3(999) As Double
    Private m_DataIndex As Int32

    Private m_ReadyToSend As Boolean = True
    Private m_DataQueue As New Queue


    Private Sub ButtonCreateTags_Click(sender As Object, e As EventArgs) Handles ButtonCreateTags.Click
        Dim TagsToAdd As New ArrayList
        TagsToAdd.Add("Tag,Value - Data Type")
        TagsToAdd.Add("Buffer Test.DateArray,ArrayString")
        TagsToAdd.Add("Buffer Test.ValueArray1,ArrayDouble")
        TagsToAdd.Add("Buffer Test.ValueArray2,ArrayDouble")
        TagsToAdd.Add("Buffer Test.ValueArray3,ArrayDouble")
        TagsToAdd.Add("Buffer Test.Ready,Boolean")
        TagsToAdd.Add("Buffer Test.Confirmation,Boolean")
        TagsToAdd.Add("Buffer Test.Error,Integer")
        m_OASConfiguration.TagCSVImport(CType(TagsToAdd.ToArray(GetType(String)), String()), TextBoxNetworkNode.Text)
    End Sub

    Private Sub ButtonRunSimulation_Click(sender As Object, e As EventArgs) Handles ButtonRunSimulation.Click
        ' Feedback variables from OAS to controller for handshaking
        Dim TagsToMonitor As New ArrayList
        TagsToMonitor.Add("Buffer Test.Confirmation.Value")
        TagsToMonitor.Add("Buffer Test.Error.Value")

        If TimerSimulation.Enabled Then
            TimerSimulation.Stop()
            m_OASData.RemoveTags(CType(TagsToMonitor.ToArray(GetType(String)), String()))
            ButtonRunSimulation.BackColor = Color.Red
            TextBoxNetworkNode.Enabled = True
        Else
            SyncLock m_DataQueue.SyncRoot
                m_DataQueue.Clear()
            End SyncLock

            m_DataIndex = 0
            m_ReadyToSend = True
            m_OASData.WriteTag("Buffer Test.Ready.Value", False)
            m_OASData.WriteTag("Buffer Test.Confirmation.Value", False)
            TextBoxNetworkNode.Enabled = False
            m_LastDate = Date.Now
            m_LastTimeFeedbackReceived = Date.Now
            If m_LastNetworkNodeUsed <> TextBoxNetworkNode.Text Then
                m_OASDataNetwork.RemoveNetworkNodeAlias(m_LastNetworkNodeUsed)
                m_LastNetworkNodeUsed = ""
            End If
            If TextBoxNetworkNode.Text.ToLower <> "localhost" Then
                m_OASDataNetwork.AddNetworkNodeAlias("localhost", TextBoxNetworkNode.Text)
                m_LastNetworkNodeUsed = TextBoxNetworkNode.Text
            End If
            TimerSimulation.Start()
            m_OASData.AddTags(CType(TagsToMonitor.ToArray(GetType(String)), String()))
            ButtonRunSimulation.BackColor = Color.Green
        End If

    End Sub

    Private Sub ButtonSetupDataLogging_Click(sender As Object, e As EventArgs) Handles ButtonSetupDataLogging.Click
        Try
            Process.Start("https://openautomationsoftware.com/knowledge-base/log-buffered-data-from-a-plc-or-controller/")
        Catch

        End Try
    End Sub

    Private Sub TimerSimulation_Tick(sender As Object, e As EventArgs) Handles TimerSimulation.Tick
        If m_InSimuation Then
            Exit Sub
        End If
        m_InSimuation = True
        Try
            ' Feedback values from OAS sent to controller
            Dim localTagValues(-1) As ClassTagValue
            SyncLock m_DataQueue.SyncRoot
                If m_DataQueue.Count > 0 Then
                    ReDim localTagValues(m_DataQueue.Count - 1)
                    m_DataQueue.CopyTo(localTagValues, 0)
                    m_DataQueue.Clear()
                End If
            End SyncLock
            Dim localTagValue As ClassTagValue
            For Each localTagValue In localTagValues
                If localTagValue.Quality Then
                    ' When controller receives feedback confirmation ready for next write to OAS
                    If localTagValue.TagId = "Buffer Test.Confirmation.Value" Then
                        If localTagValue.Value Then
                            m_OASData.WriteTag("Buffer Test.Ready.Value", False)
                            m_LastTimeFeedbackReceived = Now
                            m_ReadyToSend = True
                        Else
                            Dim temp As Int32 = 123
                        End If
                    End If
                    ' When controller receives error reset trigger and trigger alarm
                    If localTagValue.TagId = "Buffer Test.Error.Value" Then
                        If localTagValue.Value <> 0 Then
                            m_OASData.WriteTag("Buffer Test.Ready.Value", False)
                            ' Set alarm condition in controller that recording did not succeed
                            m_LastTimeFeedbackReceived = Now
                            m_ReadyToSend = True
                        End If
                    End If
                End If
            Next

            ' Simulates stacking values in the controller queue
            Dim localLastMs As Int32 = m_LastDate.Millisecond
            Dim localTime As Date = Now
            Dim localTimeDiff As TimeSpan = localTime.Subtract(m_LastDate)
            Dim localNumberOfRecordsToAdd = localTimeDiff.TotalMilliseconds
            Dim localIndex As Int32
            For localIndex = 0 To localNumberOfRecordsToAdd - 1
                m_DateQueue.Enqueue(m_LastDate.ToShortDateString + " " + m_LastDate.ToString("HH:mm:ss.fff"))
                m_ValueQueue1.Enqueue(localLastMs)
                m_ValueQueue2.Enqueue(localLastMs / 10)
                m_ValueQueue3.Enqueue(localLastMs / 100)
                m_LastDate = m_LastDate.AddMilliseconds(1)
                If localLastMs = 999 Then
                    localLastMs = 0
                Else
                    localLastMs += 1
                End If
            Next

            Dim localFeedbackTimeDiff As TimeSpan = localTime.Subtract(m_LastTimeFeedbackReceived)
            If localFeedbackTimeDiff.Seconds > 10 Then
                ' Set alarm condition in controller that OAS is not communicating
                m_ReadyToSend = True
                m_OASData.WriteTag("Buffer Test.Ready.Value", False)
                m_OASData.WriteTag("Buffer Test.Confirmation.Value", False)
            End If

            ' Move array values to OAS when arrays are full
            Do While m_DateQueue.Count > 0
                If m_DataIndex < 1000 Then
                    m_Date(m_DataIndex) = m_DateQueue.Dequeue
                    m_Value1(m_DataIndex) = m_ValueQueue1.Dequeue
                    m_Value2(m_DataIndex) = m_ValueQueue2.Dequeue
                    m_Value3(m_DataIndex) = m_ValueQueue3.Dequeue
                    m_DataIndex += 1
                Else
                    If Not (m_ReadyToSend) Then
                        Exit Do
                    End If
                End If

                If m_ReadyToSend And m_DataIndex = 1000 Then
                    m_DataIndex = 0
                    Dim localTagsToWrite As New ArrayList
                    Dim localValuesToWrite As New ArrayList
                    localTagsToWrite.Add("Buffer Test.DateArray.Value")
                    localValuesToWrite.Add(m_Date)
                    localTagsToWrite.Add("Buffer Test.ValueArray1.Value")
                    localValuesToWrite.Add(m_Value1)
                    localTagsToWrite.Add("Buffer Test.ValueArray2.Value")
                    localValuesToWrite.Add(m_Value2)
                    localTagsToWrite.Add("Buffer Test.ValueArray3.Value")
                    localValuesToWrite.Add(m_Value3)
                    localTagsToWrite.Add("Buffer Test.Confirmation.Value")
                    localValuesToWrite.Add(False)
                    localTagsToWrite.Add("Buffer Test.Ready.Value")
                    localValuesToWrite.Add(True)
                    m_OASData.WriteTags(CType(localTagsToWrite.ToArray(GetType(String)), String()), CType(localValuesToWrite.ToArray(GetType(Object)), Object()))
                    m_ReadyToSend = False
                    Exit Do
                End If
            Loop
        Catch

        End Try
        m_InSimuation = False
    End Sub

    ' OAS values
    Private Sub m_OASData_ValuesChangedAll(Tags() As String, Values() As Object, Qualities() As Boolean, TimeStamps() As Date) Handles m_OASData.ValuesChangedAll
        Dim tagIndex As Int32
        SyncLock m_DataQueue.SyncRoot
            For tagIndex = 0 To Tags.Count - 1
                m_DataQueue.Enqueue(New ClassTagValue(Tags(tagIndex), Values(tagIndex), TimeStamps(tagIndex), Qualities(tagIndex)))
            Next
        End SyncLock
    End Sub
End Class

Public Class ClassTagValue
    Private m_TagId As String
    Private m_Value As Object
    Private m_TimeStamp As DateTime
    Private m_Quality As Boolean

    Public Property TagId As String
        Get
            Return m_TagId
        End Get
        Set(value As String)
            m_TagId = value
        End Set
    End Property

    Public Property Value As Object
        Get
            Return m_Value
        End Get
        Set(value As Object)
            m_Value = value
        End Set
    End Property

    Public Property TimeStamp As DateTime
        Get
            Return m_TimeStamp
        End Get
        Set(value As DateTime)
            m_TimeStamp = value
        End Set
    End Property

    Public Property Quality As Boolean
        Get
            Return m_Quality
        End Get
        Set(value As Boolean)
            m_Quality = value
        End Set
    End Property

    Public Sub New(ByVal newTagId As String, ByVal newValue As Object, ByVal newTimeStamp As DateTime, ByVal newQuality As Boolean)
        m_TagId = newTagId
        m_Value = newValue
        m_TimeStamp = newTimeStamp
        m_Quality = newQuality
    End Sub

End Class
